static const FFCodecParser * const parser_list[] = {
    &ff_aac_parser,
    &ff_aac_latm_parser,
    &ff_ac3_parser,
    &ff_adx_parser,
    &ff_ahx_parser,
    &ff_amr_parser,
    &ff_apv_parser,
    &ff_av1_parser,
    &ff_avs2_parser,
    &ff_avs3_parser,
    &ff_bmp_parser,
    &ff_cavsvideo_parser,
    &ff_cook_parser,
    &ff_cri_parser,
    &ff_dca_parser,
    &ff_dirac_parser,
    &ff_dnxhd_parser,
    &ff_dnxuc_parser,
    &ff_dolby_e_parser,
    &ff_dpx_parser,
    &ff_dvaudio_parser,
    &ff_dvbsub_parser,
    &ff_dvdsub_parser,
    &ff_dvd_nav_parser,
    &ff_evc_parser,
    &ff_flac_parser,
    &ff_ftr_parser,
    &ff_ffv1_parser,
    &ff_g723_1_parser,
    &ff_g729_parser,
    &ff_gif_parser,
    &ff_gsm_parser,
    &ff_h261_parser,
    &ff_h263_parser,
    &ff_h264_parser,
    &ff_hevc_parser,
    &ff_hdr_parser,
    &ff_ipu_parser,
    &ff_jpeg2000_parser,
    &ff_jpegxl_parser,
    &ff_jpegxs_parser,
    &ff_lcevc_parser,
    &ff_misc4_parser,
    &ff_mjpeg_parser,
    &ff_mlp_parser,
    &ff_mpeg4video_parser,
    &ff_mpegaudio_parser,
    &ff_mpegvideo_parser,
    &ff_opus_parser,
    &ff_prores_parser,
    &ff_png_parser,
    &ff_pnm_parser,
    &ff_prores_raw_parser,
    &ff_qoi_parser,
    &ff_rv34_parser,
    &ff_sbc_parser,
    &ff_sipr_parser,
    &ff_tak_parser,
    &ff_vc1_parser,
    &ff_vorbis_parser,
    &ff_vp3_parser,
    &ff_vp8_parser,
    &ff_vp9_parser,
    &ff_vvc_parser,
    &ff_webp_parser,
    &ff_xbm_parser,
    &ff_xma_parser,
    &ff_xwd_parser,
    NULL };
