static const URLProtocol * const url_protocols[] = {
    &ff_async_protocol,
    &ff_bluray_protocol,
    &ff_cache_protocol,
    &ff_concat_protocol,
    &ff_concatf_protocol,
    &ff_crypto_protocol,
    &ff_data_protocol,
    &ff_fd_protocol,
    &ff_ffrtmpcrypt_protocol,
    &ff_ffrtmphttp_protocol,
    &ff_file_protocol,
    &ff_ftp_protocol,
    &ff_gopher_protocol,
    &ff_gophers_protocol,
    &ff_http_protocol,
    &ff_httpproxy_protocol,
    &ff_https_protocol,
    &ff_icecast_protocol,
    &ff_mmsh_protocol,
    &ff_mmst_protocol,
    &ff_md5_protocol,
    &ff_pipe_protocol,
    &ff_prompeg_protocol,
    &ff_rtmp_protocol,
    &ff_rtmpe_protocol,
    &ff_rtmps_protocol,
    &ff_rtmpt_protocol,
    &ff_rtmpte_protocol,
    &ff_rtmpts_protocol,
    &ff_rtp_protocol,
    &ff_srtp_protocol,
    &ff_subfile_protocol,
    &ff_tee_protocol,
    &ff_tcp_protocol,
    &ff_tls_protocol,
    &ff_dtls_protocol,
    &ff_udp_protocol,
    &ff_udplite_protocol,
    &ff_unix_protocol,
    &ff_libsrt_protocol,
    &ff_libssh_protocol,
    &ff_libzmq_protocol,
    &ff_ipfs_gateway_protocol,
    &ff_ipns_gateway_protocol,
    NULL };
